#  sfzen/cleaners/liquidsfz.py
#
#  Copyright 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Cleanup opcodes that LiquidSFZ does not understand.
"""
from sfzen.sfz_elems import Header, normal_opcode

SUPPORTED_OPCODES = [
	'amp_random',
	'amp_velcurve_N',
	'amp_veltrack',
	'ampeg_attack',
	'ampeg_attack_onccN',
	'ampeg_attackccN',
	'ampeg_decay',
	'ampeg_decay_curveccN',
	'ampeg_decay_onccN',
	'ampeg_decayccN',
	'ampeg_delay',
	'ampeg_delay_onccN',
	'ampeg_delayccN',
	'ampeg_hold',
	'ampeg_hold_curveccN',
	'ampeg_hold_onccN',
	'ampeg_holdccN',
	'ampeg_release',
	'ampeg_release_onccN',
	'ampeg_releaseccN',
	'ampeg_sustain',
	'ampeg_sustain_curveccN',
	'ampeg_sustain_onccN',
	'ampeg_sustainccN',
	'ampeg_vel2attack',
	'ampeg_vel2decay',
	'ampeg_vel2delay',
	'ampeg_vel2hold',
	'ampeg_vel2release',
	'ampeg_vel2sustain',
	'amplfo_delay',
	'amplfo_depth',
	'amplfo_depthccN',
	'amplfo_fade',
	'amplfo_freq',
	'amplfo_freqccN',
	'amplitude',
	'amplitude_ccN',
	'amplitude_curveccN',
	'amplitude_onccN',
	'bend_down',
	'bend_up',
	'cutoff',
	'cutoff2',
	'cutoff2_curveccN',
	'cutoff2_onccN',
	'cutoff_ccN',
	'cutoff_curveccN',
	'cutoff_onccN',
	'default_path',
	'delay',
	'delay_ccN',
	'delay_curveccN',
	'delay_onccN',
	'fil2_keycenter',
	'fil2_keytrack',
	'fil2_veltrack',
	'fil_keycenter',
	'fil_keytrack',
	'fil_veltrack',
	'fileg_attack',
	'fileg_attack_onccN',
	'fileg_attackccN',
	'fileg_decay',
	'fileg_decay_onccN',
	'fileg_decayccN',
	'fileg_delay',
	'fileg_delay_onccN',
	'fileg_delayccN',
	'fileg_depth',
	'fileg_depth_onccN',
	'fileg_depthccN',
	'fileg_hold',
	'fileg_hold_onccN',
	'fileg_holdccN',
	'fileg_release',
	'fileg_release_onccN',
	'fileg_releaseccN',
	'fileg_sustain',
	'fileg_sustain_onccN',
	'fileg_sustainccN',
	'fileg_vel2attack',
	'fileg_vel2decay',
	'fileg_vel2delay',
	'fileg_vel2depth',
	'fileg_vel2hold',
	'fileg_vel2release',
	'fileg_vel2sustain',
	'fillfo_delay',
	'fillfo_depth',
	'fillfo_depthccN',
	'fillfo_fade',
	'fillfo_freq',
	'fillfo_freqccN',
	'gain_ccN',
	'group',
	'hiccN',
	'hikey',
	'hirand',
	'hivel',
	'key',
	'label_ccN',
	'label_keyN',
	'lfoN_cutoff',
	'lfoN_cutoff_onccN',
	'lfoN_delay',
	'lfoN_delay_onccN',
	'lfoN_fade',
	'lfoN_fade_onccN',
	'lfoN_freq',
	'lfoN_freq_lfoN',
	'lfoN_freq_lfoN_onccN',
	'lfoN_freq_onccN',
	'lfoN_phase',
	'lfoN_phase_onccN',
	'lfoN_pitch',
	'lfoN_pitch_curveccN',
	'lfoN_pitch_onccN',
	'lfoN_volume',
	'lfoN_volume_onccN',
	'lfoN_wave',
	'loccN',
	'lokey',
	'loop_end',
	'loop_mode',
	'loop_start',
	'loopend',
	'loopmode',
	'loopstart',
	'lorand',
	'lovel',
	'off_by',
	'off_mode',
	'off_time',
	'offby',
	'offset',
	'offset_ccN',
	'offset_onccN',
	'offset_random',
	'on_hiccN',
	'on_loccN',
	'pan',
	'pan_ccN',
	'pan_curveccN',
	'pan_onccN',
	'pitch',
	'pitch_curveccN',
	'pitch_keycenter',
	'pitch_keytrack',
	'pitch_onccN',
	'pitch_random',
	'pitchlfo_delay',
	'pitchlfo_depth',
	'pitchlfo_depthccN',
	'pitchlfo_fade',
	'pitchlfo_freq',
	'pitchlfo_freqccN',
	'resonance',
	'resonance2',
	'resonance2_ccN',
	'resonance2_curveccN',
	'resonance2_onccN',
	'resonance_ccN',
	'resonance_curveccN',
	'resonance_onccN',
	'rt_decay',
	'sample',
	'seq_length',
	'seq_position',
	'set_ccN',
	'set_hdccN',
	'set_realccN',
	'sw_default',
	'sw_hikey',
	'sw_hilast',
	'sw_label',
	'sw_last',
	'sw_lokey',
	'sw_lolast',
	'transpose',
	'trigger',
	'tune',
	'tune_ccN',
	'tune_curveccN',
	'tune_onccN',
	'volume',
	'volume_curveccN',
	'volume_onccN',
	'xf_cccurve',
	'xf_keycurve',
	'xf_velcurve',
	'xfin_hiccN',
	'xfin_hikey',
	'xfin_hivel',
	'xfin_loccN',
	'xfin_lokey',
	'xfin_lovel',
	'xfout_hiccN',
	'xfout_hikey',
	'xfout_hivel',
	'xfout_loccN',
	'xfout_lokey',
	'xfout_lovel'
]

def clean(sfz):
	for elem, _ in sfz.walk():
		if isinstance(elem, Header):
			elem._opcodes = { key:opcode \
				for key, opcode in elem._opcodes.items() \
				if normal_opcode(key) in SUPPORTED_OPCODES }


#  end sfzen/cleaners/liquidsfz.py



#  sfzen/scripts/sfz_resample.py
#
#  Copyright 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Resamples sample files to another bitrate/depth/format.
Optionally converts stereo samples to mono.
"""
import logging, argparse
from os.path import basename
from jack import Client
try:
	from jack import JackOpenError
except ImportError:
	from jack import JackError as JackOpenError
from sfzen.resampler import SFZResampler
from sfzen import SFZ


def main():
	"""
	Entry point, defined so as to make it easy to reference from bin script.
	"""
	try:
		with Client(name = basename(__file__), no_start_server = True) as client:
			default_samplerate = client.samplerate
	except JackOpenError:
		default_samplerate = 44100

	parser = argparse.ArgumentParser()
	parser.epilog = """
	Modify the sample rate (and other properties) of the source samples of an SFZ file.
	"""
	parser.add_argument('Source', type = str, help = 'SFZ file to resample.')
	parser.add_argument('Target', type = str, nargs = '?', help = 'Target SFZ to save.')
	parser.add_argument("--rate", "-r",
		type = int, default = default_samplerate,
		help = f"Sample rate (default {default_samplerate})")
	parser.add_argument("--mono", "-m", action = "store_true",
		help = "Make samples monophonic.")
	parser.add_argument("--bitdepth", "-b",
		type = int, default = 16, help = "Bitdepth (default 16)")
	parser.add_argument("--verbose", "-v", action = "store_true",
		help = "Show more detailed debug information.")
	options = parser.parse_args()
	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "[%(filename)24s:%(lineno)3d] %(message)s"
	)

	sfz = SFZ(options.Source)
	SFZResampler(sfz,
		target_rate = options.rate,
		target_mono = options.mono,
		target_bitdepth = options.bitdepth
	).resample_as(options.Target or options.Source)


if __name__ == "__main__":
	main()

#  end sfzen/scripts/sfz_resample.py



#  sfzen/scripts/sfz_copy.py
#
#  Copyright 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Copies an .sfz to another location with multiple ways of handling samples.
"""
import os, sys, logging, argparse
from log_soso import log_error
from sfzen import (
	SFZ,
	SAMPLES_ABSPATH,
	SAMPLES_RESOLVE,
	SAMPLES_COPY,
	SAMPLES_SYMLINK,
	SAMPLES_HARDLINK
)


def main():
	p = argparse.ArgumentParser()
	p.add_argument('Source', type = str, help = 'SFZ file to copy from')
	p.add_argument('Target', type = str, nargs = '?', help = 'Destination to copy to')
	p.add_argument("--simplify", "-S", action = "store_true",
		help = 'Create <group> and <global> headers defining common opcodes.')
	group = p.add_mutually_exclusive_group()
	group.add_argument("--abspath", "-a", action = "store_true",
		help = 'Point to the original samples - absolute path')
	group.add_argument("--relative", "-r", action = "store_true",
		help = 'Point to the original samples - relative path')
	group.add_argument("--copy", "-c", action = "store_true",
		help = 'Copy samples to the "./samples" folder')
	group.add_argument("--symlink", "-s", action = "store_true",
		help = 'Create symlinks in the "./samples" folder')
	group.add_argument("--hardlink", "-l", action = "store_true",
		help = 'Hardlink samples in the "./samples" folder')
	p.add_argument("--dry-run", "-n", action = "store_true",
		help = "Do not make changes - just show what would be changed.")
	p.add_argument("--verbose", "-v", action = "store_true",
		help = "Show more detailed debug information")
	p.epilog = __doc__
	options = p.parse_args()
	if not os.path.isfile(options.Source):
		p.exit(f'"{options.Source}" is not a file')
	if not options.Target and not options.dry_run:
		p.error('<Target> is required when not --dry-run')
	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "[%(filename)24s:%(lineno)3d] %(message)s"
	)

	if options.abspath:
		samples_mode = SAMPLES_ABSPATH
	elif options.relative:
		samples_mode = SAMPLES_RESOLVE
	elif options.copy:
		samples_mode = SAMPLES_COPY
	elif options.symlink:
		samples_mode = SAMPLES_SYMLINK
	else:
		samples_mode = SAMPLES_HARDLINK

	sfz = SFZ(options.Source)
	if options.simplify:
		sfz = sfz.simplified()
	if options.dry_run:
		for sample in sfz.samples():
			sample.use_abspath()
		sfz.write(sys.stdout)
	else:
		target = options.Target
		if os.path.isdir(target):
			target = os.path.join(target, os.path.basename(options.Source))
		print(f'Copying {options.Source} to {target}')
		try:
			sfz.save_as(target, samples_mode)
		except OSError as err:
			if err.errno == 18:
				print(f'Error {err}')
				print('You probably tried to hardlink samples to a drive different from the one they are on.')
				print('\nTry another sample mode:\n')
				p.print_help()
			else:
				log_error(err)


if __name__ == '__main__':
	main()


#  end sfzen/scripts/sfz_copy.py



#  sfzen/utils/compile_syntax.py
#
#  Copyright 2024 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
import os, importlib
from numbers import Real
from pretty_repr import Repr
import yaml

VERSION_MAP = {
	None				: 'unknown',
	'SFZ v1'			: 'v1',
	'SFZ v2'			: 'v2',
	'ARIA'				: 'aria',
	'LinuxSampler'		: 'linuxsampler',
	'Cakewalk'			: 'cakewalk',
	'Cakewalk SFZ v2'	: 'cakewalk_v2',  # unimplementd by any player
}

def compile_syntax(yaml_file, target_file):
	print(f'Reading {yaml_file}...')
	with open(yaml_file) as ymlfile:
		syntax = yaml.load(ymlfile, Loader = yaml.SafeLoader)
	with open(target_file, 'w') as pyfile:
		pyfile.write('"""\nAutomatically generated by "compile_syntax.py"\n"""\n\n')
		pyfile.write('SYNTAX = ')
		Repr(syntax).write(pyfile)
		pyfile.write("\n")
	print(f'Wrote "{target_file}"')

def compile_opcodes(syntax_file, target_file):
	with open(syntax_file) as syntaxfile:
		filecontent = syntaxfile.read()
	syntax_start = filecontent.index('{')
	SYNTAX = eval(filecontent[syntax_start:])
	opcodes = {
		opcode['name']: opcode \
		for opcode in extract_opcodes(SYNTAX['categories'])
	}
	with open(target_file, 'w') as pyfile:
		pyfile.write('"""\nAutomatically generated by "compile_syntax.py"\n"""\n\n')
		pyfile.write('OPCODES = ')
		Repr(opcodes).write(pyfile)
		pyfile.write("\n")
	print(f'Wrote "{target_file}"')

def extract_opcodes(categories):
	for category in categories:
		opcodes = category.get('opcodes')
		if opcodes:
			yield from iter_opcodes(opcodes, category['name'])
		types = category.get('types')
		if types:
			yield from extract_opcodes(types)

def iter_opcodes(opcodes, category):
	for opcode in opcodes:
		yield from opcode_validator(opcode, category, None, None)

def opcode_validator(opcode, category, op_name, mod_type):
	opcode_meta = {
		'name'		: opcode['name'],
		'ver'		: version_code(opcode.get('version', 'unknown')),
		'category'	: category,
		'modulates'	: op_name,
		'mod_type'	: mod_type
	}
	extract_vdr_meta(opcode, opcode_meta)
	yield opcode_meta
	for alias in opcode.get('alias', []):
		alias_meta = {
			'name': alias['name'],
			'value': {'valid': "Alias(" + repr(opcode['name']) + ")"},
		}
		if 'version' in alias:
			alias_meta['ver'] = VERSION_MAP[alias['version']]
		else:
			alias_meta['ver'] = opcode_meta['ver']
		yield alias_meta
		if 'modulation' in alias:
			yield from extract_modulation(
				category,
				alias['modulation'].items(),
				alias['name'])
	if 'modulation' in opcode:
		yield from extract_modulation(
			category,
			opcode['modulation'].items(),
			opcode['name'])

def extract_modulation(category, items, op_name):
	for mod_type, modulations in items:
		if isinstance(modulations, list):  # some are just checkmarks
			for mod in modulations:
				yield from opcode_validator(mod, category, op_name, mod_type)

def extract_vdr_meta(opcode, opcode_meta):
	for key in ('value', 'index'):
		if key in opcode:
			if key not in opcode_meta:
				opcode_meta[key] = {}
			opcode_meta[key]['valid'] = validator(opcode[key])
			type_name = opcode[key].get('type_name')
			if type_name:
				opcode_meta[key]['type'] = type_name
			unit = opcode[key].get('unit')
			if unit:
				opcode_meta[key]['unit'] = unit

def validator(data_value):
	if 'min' in data_value:
		if 'max' in data_value:
			if not isinstance(data_value['max'], Real):
				# string value, eg "SampleRate / 2"
				return "Min(" + repr(data_value['min']) + ")"
			return "Range(" + repr(data_value['min']) + ", " + repr(data_value['max']) + ")"
		return "Min(" + repr(data_value['min']) + ")"
	if 'options' in data_value:
		return "Choice(" + repr([o['name'] for o in data_value['options']]) + ")"
	return "Any()"

def version_code(version):
	return VERSION_MAP.get(version, version.lower())

if __name__ == "__main__":
	app_dir = os.path.dirname(os.path.dirname(__file__))
	yaml_file = os.path.join(app_dir, 'res', 'syntax.yml')
	syntax_file = os.path.join(app_dir, 'syntax.py')
	opcodes_file = os.path.join(app_dir, 'opcodes.py')
	if not os.path.exists(syntax_file):
		compile_syntax(yaml_file, syntax_file)
	if not os.path.exists(opcodes_file):
		compile_opcodes(syntax_file, opcodes_file)
	module = importlib.import_module('sfzen.syntax')
	print('Successfully imported sfzen.syntax')
	module = importlib.import_module('sfzen.opcodes')
	print('Successfully imported sfzen.opcodes')

#  end sfzen/utils/compile_syntax.py
